from inspect import getdoc


def document_func_view(serializer_class=None,
                       response_serializer_class=None,
                       filter_backends=None,
                       permission_classes=None,
                       authentication_classes=None,
                       doc_format_args=list(),
                       doc_format_kwargs=dict()):
    """
    Decorator to make functional view documentable via drf-autodocs
    """
    def decorator(func):
        if serializer_class:
            func.cls.serializer_class = serializer_class
        if response_serializer_class:
            func.cls.response_serializer_class = response_serializer_class
        if filter_backends:
            func.cls.filter_backends = filter_backends
        if permission_classes:
            func.cls.permission_classes = permission_classes
        if authentication_classes:
            func.cls.authentication_classes = authentication_classes
        if doc_format_args or doc_format_kwargs:
            func.cls.__doc__ = getdoc(func).format(*doc_format_args, **doc_format_kwargs)
        return func

    return decorator


def format_docstring(*args, **kwargs):
    """
    Decorator for clean docstring formatting
    """
    def decorator(func):
        func.__doc__ = getdoc(func).format(*args, **kwargs)
        return func
    return decorator



import markdown
from rest_framework.utils.formatting import markup_description
from django import template
from django.template.defaultfilters import stringfilter
from django.utils.safestring import mark_safe, SafeData
from django.utils.text import normalize_newlines
from django.utils.html import escape
from ..endpoint import Endpoint


register = template.Library()


@register.filter(name='markdownify')
@stringfilter
def markdownify(value):
    return markup_description(value)


@register.filter(is_safe=True, needs_autoescape=True, name='keep_formatting')
@stringfilter
def keep_spacing(value, autoescape=None):
    autoescape = autoescape and not isinstance(value, SafeData)
    value = normalize_newlines(value)
    if autoescape:
        value = escape(value)
    value = mark_safe(value.replace('  ', ' &nbsp;').replace('\t', '&emsp;'))
    return mark_safe(value.replace('\n', '<br />'))


@register.filter()
def is_endpoint(obj):
    return isinstance(obj, Endpoint)


@register.filter()
def is_method_field(obj):
    return obj['type'] == 'SerializerMethodField'


@register.filter()
def add_one(value):
    data = value + 1
    return data


@register.filter()
def concat(value, s):
    try:
        return str(value) + str(s)
    except (TypeError, ValueError):
        return ""



from django.views.generic.base import TemplateView
from .parser import TreeAPIParser


class TreeView(TemplateView):

    template_name = "drf_autodocs/index.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['endpoints_tree'] = TreeAPIParser().endpoints_tree.to_dict()
        return context



from rest_framework.filters import OrderingFilter, SearchFilter
from drf_autodocs.util.rendering import render_fields_list

ordering_filter_docstring = """
Orders objects according to value of some field.
For reverse, put '-' before field name.
Example:
    'api/v1/list-of-objects?ordering=-field'
"""

search_filter_docstring = """
Allows to search items via values in some fields
If item has "^username", "=email" fields available via search,
and you do 'api/endpoint?search=email@example.com',
it will return items which email is exactly search string OR username begins with search string

By default, searches will use case-insensitive partial matches.
Options:
 '^' - starts-with search
 '=' - exact match
 '@' - full-text search
 '$' - regex search
"""

filter_backends = {
    OrderingFilter: ordering_filter_docstring,
    SearchFilter: lambda fields: search_filter_docstring + render_fields_list(*fields)
}





from django.conf import settings
from importlib import import_module

from django.utils.module_loading import import_string
from django.urls import URLResolver as RegexURLResolver
from django.urls import URLPattern as RegexURLPattern
from addict import Dict
from rest_framework.views import APIView
from .endpoint import Endpoint
from django.contrib.admindocs.views import simplify_regex


class BaseAPIParser:
    """
    Class to iherit other parsers from
    """
    def __init__(self, patterns=None):
        self.patterns = patterns
        if not patterns:
            try:
                root_urlconf = import_string(settings.ROOT_URLCONF)
            except ImportError:
                # Handle a case when there's no dot in ROOT_URLCONF
                root_urlconf = import_module(settings.ROOT_URLCONF)
            if hasattr(root_urlconf, 'urls'):
                self.patterns = root_urlconf.urls.urlpatterns
            else:
                self.patterns = root_urlconf.urlpatterns
        self.parse()

    def parse(self):
        raise NotImplementedError("Inherit some parser from this class first")

    @staticmethod
    def _is_drf_pattern(pattern):
        if hasattr(pattern.callback, 'cls'):
            return issubclass(pattern.callback.cls, APIView)


class TreeAPIParser(BaseAPIParser):
    """
    Creates a nice tree of API
    """
    def __init__(self, *args, **kwargs):
        self.endpoints_tree = Dict()
        super().__init__(*args, **kwargs)

    def parse(self):
        self.parse_tree(self.patterns, self.endpoints_tree)

    def parse_tree(self, urlpatterns, parent_node, prefix=''):
        for pattern in urlpatterns:
            if isinstance(pattern, RegexURLResolver):
                try:
                    regex = pattern._regex if hasattr(pattern, "_regex") else pattern.pattern._regex
                except:
                    regex = ""
                child_node_name = simplify_regex(regex).strip('/') if regex else ""
                self.parse_tree(
                    urlpatterns=pattern.url_patterns,
                    parent_node=parent_node[child_node_name] if child_node_name else parent_node,
                    prefix='%s/%s' % (prefix, child_node_name)
                )

            elif isinstance(pattern, RegexURLPattern) and self._is_drf_pattern(pattern):
                api_endpoint = Endpoint(pattern, prefix)
                parent_node[api_endpoint.name] = api_endpoint










from rest_framework.serializers import BaseSerializer, ChoiceField, RelatedField, ManyRelatedField
from rest_framework.filters import SearchFilter
from inspect import getdoc
from django.contrib.admindocs.views import simplify_regex
from drf_autodocs import builtin_docs
from django.conf import settings


class Endpoint:
    counter = 0

    def __init__(self, pattern, prefix=None):
        self.id = Endpoint.counter
        Endpoint.counter += 1

        self.pattern = pattern
        self.view = pattern.callback
        self.methods = self._get_allowed_methods()
        self.complete_path = self._get_complete_path(pattern, prefix)

        self.name = self._get_endpoint_name()

        if hasattr(self.view.cls, 'extra_url_params'):
            self.extra_url_params = self.view.cls.extra_url_params

        if hasattr(self.view.cls, 'filter_backends') and len(getattr(self.view.cls, 'filter_backends')) > 0:
            self._collect_filter_backends()

        if hasattr(self.view.cls, 'authentication_classes') and self.view.cls.authentication_classes is not None:
            self.authentication_classes = [(cls.__name__, getdoc(cls)) for cls in self.view.cls.authentication_classes]

        if hasattr(self.view.cls, 'permission_classes') and self.view.cls.permission_classes is not None:
            self.permission_classes = [(cls.__name__, getdoc(cls)) for cls in self.view.cls.permission_classes]

        self.docstring = self._get_doc()

        if hasattr(self.view.cls, 'serializer_class') and self.view.cls.serializer_class is not None:
            if not set(self.methods) == {'GET', 'OPTIONS'}:
                self.input_fields = self._get_serializer_fields(self.view.cls.serializer_class())
            else:
                self.output_fields = self._get_serializer_fields(self.view.cls.serializer_class())

        if hasattr(self.view.cls, 'response_serializer_class'):
            self.output_fields = self._get_serializer_fields(self.view.cls.response_serializer_class())

    def _get_doc(self):
        no_description = "No description provided by developer"
        try:
            return self.view.cls.__doc__ if self.view.cls.__doc__ is not None else no_description
        except AttributeError:
            return no_description

    def _get_endpoint_name(self):
        if hasattr(settings, 'AUTODOCS_ENDPOINT_NAMES') and settings.AUTODOCS_ENDPOINT_NAMES == 'view':
            ret = ''.join(
                [
                    (' %s' % c if c.isupper() and not self.view.__name__.startswith(c) else c)
                    for c in self.view.__name__
                    ]
            ).replace('-', ' ').replace('_', ' ').title()
            return ret
        else:
            return self.pattern.name.replace('-', ' ').replace('_', ' ').title()

    def _collect_filter_backends(self):
        self.filter_backends = []
        for f in self.view.cls.filter_backends:
            if f in builtin_docs.filter_backends:
                if f is SearchFilter:
                    if hasattr(self.view.cls, 'search_filters'):
                        doc = builtin_docs.filter_backends[f](self.view.cls.search_filters)
                    else:
                        doc = "Developer didn't specify any fields for search"
                else:
                    doc = builtin_docs.filter_backends[f]
                self.filter_backends.append((f.__name__, doc))
            else:
                self.filter_backends.append((f.__name__, getdoc(f)))

    def _get_allowed_methods(self):
        if hasattr(self.view, 'cls'):
            return [m.upper() for m in self.view.cls.http_method_names if hasattr(self.view.cls, m)]
        else:
            return []

    @staticmethod
    def _get_complete_path(pattern, prefix=None):
        regex = pattern._regex if hasattr(pattern, "_regex") else pattern.pattern._regex
        return prefix + simplify_regex(regex)

    def _get_serializer_fields(self, serializer):
        fields = []

        if hasattr(serializer, 'get_fields'):
            for key, field in serializer.get_fields().items():
                to_many_relation = True if hasattr(field, 'many') else False
                sub_fields = []

                if to_many_relation:
                    sub_fields = self._get_serializer_fields(field.child) if isinstance(field, BaseSerializer) else None
                else:
                    sub_fields = self._get_serializer_fields(field) if isinstance(field, BaseSerializer) else None
                field_data = {
                    "name": key,
                    "read_only": field.read_only,
                    "type": str(field.__class__.__name__),
                    "sub_fields": sub_fields,
                    "required": field.required,
                    "to_many_relation": to_many_relation,
                    "help_text": field.help_text,
                    "write_only": field.write_only
                }
                if isinstance(field, ChoiceField) and not isinstance(field, (RelatedField, ManyRelatedField)):
                    field_data['choices'] = field.choices

                if isinstance(field, RelatedField):
                    if hasattr(field, 'queryset') and hasattr(field.queryset, 'model'):
                        field_data['help_text'] = ('{}\nRequires/renders pk(id) of {} as integer'.format(
                            field.help_text if field.help_text else "",
                            field.queryset.model.__name__)
                        )
                    elif hasattr(serializer.Meta.model, key):
                        field_data['help_text'] = ('{}\nRequires/renders pk(id) of {} as integer'.format(
                            field.help_text if field.help_text else "",
                            getattr(serializer.Meta.model, key).field.related_model.__name__)
                        )
                elif isinstance(field, ManyRelatedField):
                    if hasattr(field, 'queryset') and hasattr(field.queryset, 'model'):
                        field_data['help_text'] = ("{}\nRequires/renders list of pk's(id's) of {} objects.".format(
                            field.help_text if field.help_text else "",
                            field.child_relation.queryset.model.__name__)
                        )
                    elif hasattr(serializer.Meta.model, key):
                        field_data['help_text'] = ('{}\nRequires/renders pk(id) of {} as integer'.format(
                            field.help_text if field.help_text else "",
                            getattr(serializer.Meta.model, key).field.related_model.__name__)
                        )

                fields.append(field_data)

        return fields








from setuptools import find_packages, setup

setup(
    name="drf_autodocs",
    version=__import__('drf_autodocs').__version__,
    author="Mashianov Oleksander",
    author_email="mashianov@gmail.com",
    packages=find_packages(),
    include_package_data=True,
    url="https://github.com/iMakedonsky/drf-autodocs",
    license='BSD',
    description="Extensible auto django rest framework api generator",
    long_description=open("README.md").read(),
    install_requires=[
        'django>=1.11',
        'markdown>=2.6.7',
        'addict>=2.0.0',
        'djangorestframework>=3.7'
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
)
